/*	$LAAS$ */
/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include <portLib.h>
#include <errnoLib.h>

#include <neobotix/neo.h>
#include <neobotix/pscan.h>
#include <neobotix/canMotors.h>

#include "canIds.h"
#include "harmonica.h"

/*----------------------------------------------------------------------*/
/*
 * Initialize communication with the motor controllers 
 */
CAN_MOTORS_STR *
canMotorsInit(CAN_DEV *dev)
{
	CAN_MOTORS_STR *motors;
	CAN_HARMONICA_STR *m;
	int i;

	motors = (CAN_MOTORS_STR *)malloc(sizeof(CAN_MOTORS_STR));
	if (motors == NULL) {
		errnoSet(S_canMotorsLib_NO_MEMORY);
		return NULL;
	}
	motors->mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	if (motors->mutex == NULL) {
		errnoSet(S_canMotorsLib_NO_MEMORY);
		goto fail;
	}
	pthread_mutex_init(motors->mutex, NULL);

	for (i = 0; i < CAN_MOTORS_N_MOTORS; i++) {
		/* Initialize each controller */
		m = harmonicaInit(dev, i);
		if (m == NULL) {
			goto fail;
		}
		motors->motor[i] = m;
		m->parent = motors;
		if (harmonicaInitCtrl(m) == ERROR) {
			goto fail;
		}
	} /* for */
	motors->status = OK;
	return motors;
fail:
	for (i = 0; i < CAN_MOTORS_N_MOTORS; i++) {
		if (motors->motor[i] != NULL) {
			harmonicaEnd(motors->motor[i]);
			free(motors->motor[i]);
			motors->motor[i] = NULL;
		}
	}
	if (motors->mutex != NULL) {
		pthread_mutex_destroy(motors->mutex);
		free(motors->mutex);
	}
	free(motors);
	return NULL;
}

/*----------------------------------------------------------------------*/
/*
 * End communication with the motors controllers 
 */
void
canMotorsEnd(CAN_MOTORS_STR *motors)
{
	int i;

	for (i = 0; i < CAN_MOTORS_N_MOTORS; i++) {
		if (motors->motor[i] != NULL) {
			harmonicaEnd(motors->motor[i]);
			motors->motor[i] = NULL;
		}
	}
	if (motors->mutex != NULL) {
		pthread_mutex_destroy(motors->mutex);
		free(motors->mutex);
	}
	free(motors);
}
	
/*----------------------------------------------------------------------*/
/*
 * Start the motor controllers 
 */
STATUS
canMotorsStart(NEO_STR *neo)
{
	CAN_MOTORS_STR *motors = neo->motors;
	int i;
	CAN_HARMONICA_STR *m;

	pthread_mutex_lock(motors->mutex);
	for (i = 0; i < CAN_MOTORS_N_MOTORS; i++) {
		m = motors->motor[i];
		if (m != NULL) {
			if (harmonicaStartWatchdog(m, TRUE) == ERROR) {
				pthread_mutex_unlock(motors->mutex);
				return ERROR;
			}
			if (harmonicaStart(m) == ERROR) {
				pthread_mutex_unlock(motors->mutex);
				return ERROR;
			}
			m->angleGearRadMem = 0;
			harmonicaSetGearVelRadS(m, 0);
		}
	} /* for */
	pthread_mutex_unlock(motors->mutex);
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Reset .the motor controllers 
 */
STATUS
canMotorsReset(NEO_STR *neo)
{
	CAN_MOTORS_STR *motors = neo->motors;
	int i;
	CAN_HARMONICA_STR *m;

	pthread_mutex_lock(motors->mutex);
	for (i = 0; i < CAN_MOTORS_N_MOTORS; i++) {
		m = motors->motor[i];
		if (m != NULL) {
			if (harmonicaReset(m) == ERROR) {
				pthread_mutex_unlock(motors->mutex);
				return ERROR;
			}
			harmonicaSetGearVelRadS(m, 0);
		}
	} /* for */
	pthread_mutex_unlock(motors->mutex);
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Set the speed of a motor
 */
void 
canMotorsSetVelGearRadS(NEO_STR *neo, double lvel, double rvel)
{
	CAN_MOTORS_STR *motors = neo->motors;
	if (motors->status != OK) {
		rvel = lvel = 0;
	}
	pthread_mutex_lock(motors->mutex);
	harmonicaSetGearVelRadS(motors->motor[CAN_MOTOR_LEFT], lvel);
	harmonicaSetGearVelRadS(motors->motor[CAN_MOTOR_RIGHT], rvel);
	pthread_mutex_unlock(motors->mutex);
}

/*----------------------------------------------------------------------*/
/*
 * Returns current position (?) and speed of a motor
 */
void 
canMotorsGetGearPosVelRadS(NEO_STR *neo, double *lpos, double *lvel, 
    double *rpos, double *rvel)
{
	CAN_MOTORS_STR *motors = neo->motors;

	pthread_mutex_lock(motors->mutex);
	*lpos = motors->motor[CAN_MOTOR_LEFT]->posGearMeasRad;
	*lvel = motors->motor[CAN_MOTOR_LEFT]->velGearMeasRadS;
	*rpos = motors->motor[CAN_MOTOR_RIGHT]->posGearMeasRad;
	*rvel = motors->motor[CAN_MOTOR_RIGHT]->velGearMeasRadS;
	pthread_mutex_unlock(motors->mutex);
}
/*----------------------------------------------------------------------*/
/*
 * Returns relative motion and speed of a motor since last call
 */
void 
canMotorsGetGearDeltaPosVelRadS(NEO_STR *neo, double *lDeltaPos, double *lVel,
    double *rDeltaPos, double *rVel)
{
	CAN_MOTORS_STR *motors = neo->motors;

	pthread_mutex_lock(motors->mutex);
	*lDeltaPos = motors->motor[CAN_MOTOR_LEFT]->posGearMeasRad 
	    - motors->motor[CAN_MOTOR_LEFT]->angleGearRadMem;
	*lVel = motors->motor[CAN_MOTOR_LEFT]->velGearMeasRadS;
	*rDeltaPos = motors->motor[CAN_MOTOR_RIGHT]->posGearMeasRad 
	    - motors->motor[CAN_MOTOR_RIGHT]->angleGearRadMem;
	*rVel = motors->motor[CAN_MOTOR_RIGHT]->velGearMeasRadS;
	motors->motor[CAN_MOTOR_LEFT]->angleGearRadMem = 
	    motors->motor[CAN_MOTOR_LEFT]->posGearMeasRad;
	motors->motor[CAN_MOTOR_RIGHT]->angleGearRadMem = 
	    motors->motor[CAN_MOTOR_RIGHT]->posGearMeasRad;
	pthread_mutex_unlock(motors->mutex);
}

/*----------------------------------------------------------------------*/
/*
 * Returns the operational status of one motor 
 */
CAN_MOTOR_STATE 
canMotorsState(NEO_STR *neo, CAN_MOTOR_ID id)
{
	CAN_MOTORS_STR *motors = neo->motors;
	CAN_MOTOR_STATE ret;

	pthread_mutex_lock(motors->mutex);
	ret = motors->motor[id]->motorState;
	pthread_mutex_unlock(motors->mutex);
	return ret;
}

/*----------------------------------------------------------------------*/
/*
 * Show motors status
 */
void
canMotorsShow(NEO_STR *neo)
{
	CAN_MOTORS_STR *motors = neo->motors;
	int i;

	for (i = 0; i < CAN_MOTORS_N_MOTORS; i++) {
		printf("-- motor %d\n", i);
		harmonicaShow(motors->motor[i]);
	}
	printf("--\n");
}
	   
/*----------------------------------------------------------------------*/
/*
 * Decode CAN messages coming from motor controllers 
 */
BOOL
canMotorsDecode(CAN_MOTORS_STR *motors, CAN_MSG *m)
{
	BOOL ret;

	pthread_mutex_lock(motors->mutex);

	ret = harmonicaDecode(motors->motor[CAN_MOTOR_LEFT], m);
	ret |= harmonicaDecode(motors->motor[CAN_MOTOR_RIGHT], m);

	pthread_mutex_unlock(motors->mutex);
	return ret;
}
