/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ident "$Id$"


#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errnoLib.h>
#include <portLib.h>
#include <h2errorLib.h>

#include <neobotix/pscan.h>
#include <neobotix/neo.h>
#include <neobotix/canUS.h>

#include "canUSCmd.h"
#include "canIds.h"

/*----------------------------------------------------------------------*/
/*
 * Initialize communication with the US board
 */
CAN_US_STR *
canUSInit(CAN_DEV *dev)
{
	CAN_US_STR *us;

	us = (CAN_US_STR *)malloc(sizeof(CAN_US_STR));
	if (us == NULL) {
		errnoSet(S_canUsLib_NO_MEMORY);
		return NULL;
	}
	us->mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	if (us->mutex == NULL) {
		free(us);
		errnoSet(S_canUsLib_NO_MEMORY);
		return NULL;
	}
	pthread_mutex_init(us->mutex, NULL);
	us->dev = dev;
	us->recMsgId = USBOARD_RX_ID;
	us->scaleToMM = 0.543;
	return us;
}

/*----------------------------------------------------------------------*/
/*
 * Ends communication with the US board
 */
void 
canUSEnd(CAN_US_STR *us)
{
	pthread_mutex_destroy(us->mutex);
	free(us->mutex);
	free(us);
}

/*----------------------------------------------------------------------*/
/* 
 * Send the 'connect' message to the US board
 */
void
canUSRequestConnect(NEO_STR *neo)
{
	CAN_US_STR *us = neo->us;
	CAN_MSG m;

	canMsgSet(&m, USBOARD_TX_ID, 
	    CMD_USBOARD_CONNECT, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(us->dev, &m);
}

/*----------------------------------------------------------------------*/
/* 
 * Make the sensors start shooting
 */
void
canUSStart(NEO_STR *neo)
{
	CAN_US_STR *us = neo->us;
	CAN_MSG m;

	canMsgSet(&m, USBOARD_TX_ID,
	    CMD_USBOARD_USSTART, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(us->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Stop shooting
 */
void
canUSStop(NEO_STR *neo)
{
	CAN_US_STR *us = neo->us;
	CAN_MSG m;

	canMsgSet(&m, USBOARD_TX_ID,
	    CMD_USBOARD_USSTOP, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(us->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Ask the US board for measures (only the 1st 8 sensors)
 */
void
canUSRequestDist(NEO_STR *neo)
{
	CAN_US_STR *us = neo->us;
	CAN_MSG m;

	canMsgSet(&m, USBOARD_TX_ID, 
	    CMD_USBOARD_GET1TO8, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(us->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Ask the US board for analog sensors data (not connected)
 */
void
canUSRequestAnalog(NEO_STR *neo)
{
	CAN_US_STR *us = neo->us;
	CAN_MSG m;

	canMsgSet(&m, USBOARD_TX_ID,
	    CMD_USBOARD_GETANALOGIN, 16, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(us->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Decode CAN messages sent by the US board
 */
BOOL
canUSDecode(CAN_US_STR *us, CAN_MSG *m)
{
	int i, cmd = canGetCmd(m);
	
	if (m->id != us->recMsgId) 
		return FALSE;
	pthread_mutex_lock(us->mutex);
	canMsgShow(m);
	switch (cmd) {
	case CMD_USBOARD_CONNECT:
		if (m->data[0] == 1 && m->data[1] == 2
		    && m->data[2] == 3 && m->data[3] == 4
		    && m->data[4] == 5 && m->data[5] == 6
		    && m->data[6] == 7)
			us->connected = TRUE;
		else 
			us->connected = FALSE;
		break;
	case CMD_USBOARD_REC1TO4:
		/* US sensors 1 to 4 */
		us->dist[0] = m->data[0] | ((m->data[4] & 0xf0) << 4);
		us->dist[1] = m->data[1] | ((m->data[4] & 0x0f) << 8);
		us->dist[2] = m->data[2] | ((m->data[5] & 0xf0) << 4);
		us->dist[3] = m->data[3] | ((m->data[5] & 0x0f) << 8);
		for (i = 0; i < 4; i++) 
			us->dist[i] = us->dist[i] * us->scaleToMM;
		break;
	case CMD_USBOARD_REC5TO8:
		/* US sensors 5 to 8 */
		us->dist[4] = m->data[0] | ((m->data[4] & 0xf0) << 4);
		us->dist[5] = m->data[1] | ((m->data[4] & 0x0f) << 8);
		us->dist[6] = m->data[2] | ((m->data[5] & 0xf0) << 4);
		us->dist[7] = m->data[3] | ((m->data[5] & 0x0f) << 8);
		for (i = 4; i < 8; i++) 
			us->dist[i] = us->dist[i] * us->scaleToMM;
		break;
	case CMD_USBOARD_GETANALOGIN:
		/* Analog sensors */
		if (m->data[5] == 0) {
			us->analogIn[0] = ( m->data[0] 
			    | ((m->data[4] & 0x03) << 8)) & 0x3ff;
			us->analogIn[1] = ( m->data[1] 
			    | ((m->data[4] & 0x0c) << 6)) & 0x3ff;
			us->analogIn[2] = ( m->data[2] 
			    | ((m->data[4] & 0x30) << 4)) & 0x3ff;
			us->analogIn[3] = ( m->data[3] 
			    | ((m->data[4] & 0xc0) << 2)) & 0x3ff;
		} else if (m->data[5] == 1) {
			us->analogIn[4] = ( m->data[0] 
			    | ((m->data[4] & 0x03) << 8)) & 0x3ff;
			us->analogIn[5] = ( m->data[1] 
			    | ((m->data[4] & 0x0c) << 6)) & 0x3ff;
			us->analogIn[6] = ( m->data[2] 
			    | ((m->data[4] & 0x30) << 4)) & 0x3ff;
			us->analogIn[7] = ( m->data[3] 
			    | ((m->data[4] & 0xc0) << 2)) & 0x3ff;
		} else {
			printf("Bad CMD_USBOARD_GETANALOGIN packet\n");
			canMsgShow(m);
		}
		break;
	default:
		printf("Unkown packet\n");
		canMsgShow(m);
		break;
	}
	pthread_mutex_unlock(us->mutex);
	return TRUE;
}

/*----------------------------------------------------------------------*/
/*
 * Return an array containing the measured distances on the sensors 
 */
STATUS
canUSGetDist(NEO_STR *neo, int *dist)
{
	CAN_US_STR *us = neo->us;

	pthread_mutex_lock(us->mutex);
	memcpy(dist, us->dist, CAN_US_N_US*sizeof(int));
	pthread_mutex_unlock(us->mutex);
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Return an array of the measured values on the analog sensors 
 */
STATUS
canUSGetAnalog(NEO_STR *neo, int *ana)
{
	CAN_US_STR *us = neo->us;

	pthread_mutex_lock(us->mutex);
	memcpy(ana, us->analogIn, CAN_US_N_ANALOG*sizeof(int));
	pthread_mutex_unlock(us->mutex);
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Show the current US data on stdout
 */
void
canUSShow(NEO_STR *neo)
{
	CAN_US_STR *us = neo->us;
	int i;

	pthread_mutex_lock(us->mutex);
	printf("Connected: %s\n", us->connected ? "TRUE" : "FALSE");
	printf("Distances (mm):");
	for (i = 0; i < CAN_US_N_US; i++) {
		printf(" %d", us->dist[i]);
	}
	printf("\nAnalogIn:");
	for (i = 0; i < CAN_US_N_ANALOG; i++) {
		printf(" %d", us->analogIn[i]);
	}
	printf("\n");
	pthread_mutex_unlock(us->mutex);
}

