/*	$LAAS$ */
/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include <errnoLib.h>
#include <portLib.h>
#include <taskLib.h>

#include <neobotix/pscan.h>
#include <neobotix/neo.h>
#include <neobotix/canGyro.h>
#include <neobotix/canIO.h>
#include <neobotix/canUS.h>
#include <neobotix/canMotors.h>

static void neoReadTask(NEO_STR *);

/*----------------------------------------------------------------------*/
/*
 * Init communication with the neobotix platform
 */
NEO_STR *
neoInit(char *device)
{
	NEO_STR *n;
	CAN_MSG m;

	n = (NEO_STR *)malloc(sizeof(NEO_STR));
	if (n == NULL) {
		errnoSet(S_neoLib_NO_MEMORY);
		return NULL;
	}
	n->can = pscanInit(device);
	if (n->can == NULL) 
		return NULL;

	n->gyro = canGyroInit(n->can);
	n->io = canIOInit(n->can);
	n->us = canUSInit(n->can);
	n->motors = canMotorsInit(n->can);
	n->dying = FALSE;

	n->readTask = taskSpawn(NEO_READ_TASK_NAME,
	    NEO_READ_TASK_PRIORITY,
	    VX_FP_TASK,
	    NEO_READ_TASK_STACK_SIZE,
	    (FUNCPTR)neoReadTask, n, NULL);
	if (n->readTask == ERROR) {
		if (n->gyro) 
			canGyroEnd(n->gyro);
		pscanEnd(n->can);
		return NULL;
	}
	/* Send init message */
	canMsgSet(&m, 0, 1, 0, 0, 0, 0, 0, 0, 0);
	m.len = 2;
	pscanTransmitMsg(n->can, &m);
	return n;
}
/*----------------------------------------------------------------------*/
/*
 * End communication with the neobotix platform
 */
void
neoEnd(NEO_STR *neo)
{
	neo->dying = TRUE;
	if (neo->gyro != NULL)
		canGyroEnd(neo->gyro);
	if (neo->io != NULL)
		canIOEnd(neo->io);
	if (neo->us != NULL)
		canUSEnd(neo->us);
	if (neo->motors != NULL)
		canMotorsEnd(neo->motors);
	pscanEnd(neo->can);
	free(neo);
}

/*----------------------------------------------------------------------*/
/*
 * Body of the neoRead task, that reads and decodes CAN messages sent 
 * by the various components of the platform 
 */
static void
neoReadTask(NEO_STR *neo)
{
	CAN_MSG m;
	int n;
	BOOL decoded = FALSE;

	do {
		n = pscanReceiveMsgWait(neo->can, &m, -1);
		if (neo->dying == TRUE) 
			break;
		decoded = FALSE;
		if (neo->gyro != NULL) {
			decoded |= canGyroDecode(neo->gyro, &m);
		}
		if (neo->io != NULL) {
			decoded |= canIODecode(neo->io, &m);
		}
		if (neo->us != NULL) {
			decoded |= canUSDecode(neo->us, &m);
		}
		if (neo->motors != NULL) {
			decoded |= canMotorsDecode(neo->motors, &m);
		}
		if (!decoded) {
			fprintf(stderr, "%s: undecoded packet\n", __func__);
			canMsgShow(&m);
		}
	} while (n != ERROR);
}
